/* 
 * MPTclMain.c -- (c) 2003 Cyril BARABTO 
 *
 *	This file contains code to implement the "packer"
 *	media tcl manager for Tcl/Tk.
 *
 * tcl/tk (c) 1990-2003 The Regents of the University of California.
 * 	  (c) 1994-2003 Sun Microsystems, Inc.
 *
 * DISCLAIMER OF ALL WARRANTIES.
 *
 */
#include "MPTclInit.h"

/*
 *--------------------------------------------------------------
 * 
 * declare type def Media.
 *
 *--------------------------------------------------------------
 */


typedef struct {

	TkWindow	*mediawin;		/* widget tk windows */
	char		*mediawinName;		/* widget name */
 	Tcl_Interp 	*interp;		/* interpreter associated with widget */
	Tcl_Command 	widgetCmd;		/* token for Media widget command */

	Display 	*display;		/* display containing widget */
	Screen		*screen;		
	Window		root;
	Window		mpwin;			/* current mplayer window */
	FILE		*mpcmd;			/* current mplayer cmd line */

	int		state;			/* state currently */ 	
	
	char		*file;			/* name of movie file or list of movie files */
	char		*playlist;		/* name of play list file */
	int		dvd;			/* dvd number */

	int		width;			/* width size	*/
	int		height;			/* heigth size 	*/
	int		scale;			/* scale */
	int		background;		/* background color */

	int		volume;			/* sound volume */

	char		*args;			/* another args string */

} Media ;


/*
 *--------------------------------------------------------------
 *
 * define state
 *
 *--------------------------------------------------------------
 */

#define MPTCL_STATE_VOID		0
#define MPTCL_STATE_LOADNOTIFY		1
#define MPTCL_STATE_LOADED		2
#define MPTCL_STATE_PAUSED		4

char *mediaState[] = { "stop", "loading" , "playing" , "ERROR", "paused" };

/*
 *--------------------------------------------------------------
 *
 * define commands and options index.
 *
 *--------------------------------------------------------------
 */

/* options */
#define MPTCL_WIDTH		100 	/* w h scale options */
#define MPTCL_HEIGHT		110
#define MPTCL_SCALE		120
#define MPTCL_BACKGROUND	130
#define MPTCL_FILE		200	/* file option */
#define MPTCL_PLAYLIST		210	/* play list file option */
#define MPTCL_DVD		220	/* dvd option */
/* cmds */
#define MPTCL_CONFIG		1000	/* general configure command */
#define MPTCL_CGET		1100	/* get config comamnd */
#define MPTCL_PAUSE		1200	
#define MPTCL_VOLUME		1300
#define MPTCL_UP		1310	/* sub cmd */
#define MPTCL_DOWN		1320
#define MPTCL_STOP		1400
#define MPTCL_GET		1500
#define MPTCL_STATE		1600
/* internales cmds */
#define MPTCL_MPWIN_SET	   	9000	
#define MPTCL_MPWIN_RESIZE 	9100	
 

MPTcl_Equ mediaEqu[] = {
	{"-background",	MPTCL_BACKGROUND},
	{"-dvd", 	MPTCL_DVD},
	{"-file", 	MPTCL_FILE},
	{"-files", 	MPTCL_FILE},
	{"-height",	MPTCL_HEIGHT},
	{"-playlist", 	MPTCL_PLAYLIST},
	{"-scale", 	MPTCL_SCALE},
	{"-width", 	MPTCL_WIDTH},
	{"cget",	MPTCL_CGET},
	{"configure",	MPTCL_CONFIG},
	{"down",	MPTCL_DOWN},
	{"get",		MPTCL_GET},
	{"pause",	MPTCL_PAUSE},
	{"stop", 	MPTCL_STOP},
	{"toggle",	MPTCL_PAUSE},
	{"state",	MPTCL_STATE},
	{"up",		MPTCL_UP},
	{"volume",	MPTCL_VOLUME},
	{"_MPWIN_SET_",	MPTCL_MPWIN_SET},
	{"_MPWIN_RESIZE_", MPTCL_MPWIN_RESIZE},
	{0, 0}
};

/*
 *--------------------------------------------------------------
 *
 * declare prototypes widget Media command and proc
 *
 *--------------------------------------------------------------
 */

/* widget proc */
static int  MediaWidgetCmd(ClientData, Tcl_Interp	*, int , char **); 	/* widget cmd Media */
static void MediaWidgetDelete(ClientData); 					/* proc kill Media */
static int  MediaWidgetConfigure(Tcl_Interp *, Media *, int , char **);		/* save option and config Media widget */

/* X manipulation proc */
static Window XFindWindowId(Display *, Window , char *);			/* Find x id by name */ 

/*
 *--------------------------------------------------------------
 *
 * MPTcl_MediaCmd --
 *
 *	This procedure is invoked to process the "media" Tcl command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *--------------------------------------------------------------
 */
int
MPTcl_MediaCmd(clientData, interp, argc, argv)
    ClientData clientData;	/* Media window associated with interpreter. 	*/
    Tcl_Interp *interp;		/* Current interpreter. 			*/
    int   	argc;		/* Number of arguments. 			*/
    char 	**argv;		/* Argument strings. 				*/
    {

	TkWindow 	*tkwin	= (TkWindow *) clientData;	/* current tkwin pointer */
	Media		*mediaPtr;				/* current Media pointer on Media register */
	char		*tamponStr;
	int		error 	= TCL_OK;		

	if ((argc <2 ) || (argv[1][0]!='.')) {
		 sprintf(interp->result, "Main ?command? ?arg?...?args?!");
		return TCL_ERROR ;
	};


	/* create and init new Media option register  */
	mediaPtr 		= (Media *) malloc (sizeof(Media));
	mediaPtr->width		= 160;		/* args default */
	mediaPtr->height	= 120;
	mediaPtr->scale		= 1;
	mediaPtr->background	= 0;
	mediaPtr->volume	= 0x7F;
	mediaPtr->file		= (char *) NULL;
	mediaPtr->playlist	= (char *) NULL;
	mediaPtr->dvd		= 0;
	mediaPtr->args		= (char *) NULL;
	mediaPtr->mediawinName 	=	(char *) malloc (strlen(argv[1])+1);
	strncpy(mediaPtr->mediawinName, argv[1], strlen(argv[1])+1);
	mediaPtr->mediawin 	= (TkWindow *) Tk_CreateWindowFromPath(interp, (Tk_Window) tkwin, argv[1], NULL);
 	mediaPtr->display 	= Tk_Display(mediaPtr->mediawin);
    	mediaPtr->interp = interp;
   	mediaPtr->widgetCmd 	= Tcl_CreateCommand(interp, Tk_PathName(mediaPtr->mediawin),
						MediaWidgetCmd, (ClientData) mediaPtr,
						MediaWidgetDelete);
	mediaPtr->screen	= DefaultScreen(mediaPtr->display);
	mediaPtr->root		= DefaultRootWindow(mediaPtr->display);
	mediaPtr->mpwin		= 0;			/* configured later */
	mediaPtr->state		= MPTCL_STATE_VOID;

	/* CONFIGURE WIDGET */
	error =	MediaWidgetConfigure(interp, mediaPtr, argc-2, argv+2);

	/* config Media widget tk class */
	TkSetClassProcs((Tk_Window) mediaPtr->mediawin, NULL, (ClientData) mediaPtr);
 	Tk_SetClass((Tk_Window) mediaPtr->mediawin, "Media");

		/* free */	
	if (error == TCL_ERROR) free(mediaPtr);


	/* return error */
	return error;
};

/*
 *--------------------------------------------------------------
 *
 * MediaWidgetCmd --
 *
 * 	called by widget
 *
 * Results:
 *	A standard Tcl result.
 *
 *--------------------------------------------------------------
 */
static int
MediaWidgetCmd(clientData, interp, argc, argv)
    ClientData  clientData;	/* Media window. */
    Tcl_Interp 	*interp;	/* Current interpreter. */
    int 	argc;		/* Number of arguments. */
    char 	**argv;		/* Argument strings. */
{
   	Media	*mediaPtr = (Media *) clientData;	/* find mediaPtr */
	int 	index 		= 0;			/* index of command */
	int 	error		= TCL_OK;
	Window	win;					// CURRENTLY DEVEL
	int	x,y;
	unsigned int w, h, dw, dph;
	

	/* test args number */
	if (argc<2) {
		sprintf(interp->result, "%s ?command? ?arg?.....?arg?!", argv[0]);
		return TCL_ERROR;
	};

	/* test if mplayer is allways open */	
	if (mediaPtr->state == MPTCL_STATE_LOADED) {
		if (XFindWindowId(mediaPtr->display, mediaPtr->root, mediaPtr->mediawinName)==0) {
			/* quit mplayer */
			fprintf(mediaPtr->mpcmd, "quit\n");
			fflush(mediaPtr->mpcmd);
			pclose(mediaPtr->mpcmd); 
			/* reinit media */
			mediaPtr->state = MPTCL_STATE_VOID;
			mediaPtr->mpwin	= 0;
		};
	};

	/* indentify command */
	index = MPTcl_GetEqu(argv[1], mediaEqu);
	sprintf(interp->result, "%s command %s not exist!", argv[0], argv[1]);

	switch (index) {
		case MPTCL_CONFIG : 
			error = MediaWidgetConfigure(interp, mediaPtr, argc-2, argv+2);
			break ;
		case MPTCL_CGET : 
			if (argv[2] == NULL) break ;
			index = MPTcl_GetEqu(argv[2], mediaEqu);
			switch (index) {
				case MPTCL_BACKGROUND :
					sprintf(interp->result, "%d",mediaPtr->background);
					break;
				default :
					error = TCL_ERROR;
			};
			break ;
		case MPTCL_GET : 
			if (argv[2] == NULL) break ;
			index = MPTcl_GetEqu(argv[2], mediaEqu);
			switch (index) {
				case MPTCL_STATE :
					sprintf(interp->result, "%s" , mediaState[(mediaPtr->state>3) ? 4 : mediaPtr->state]);
					break;
				default :
					error = TCL_ERROR;
			};
			break ;
		case MPTCL_PAUSE : 
			if (mediaPtr->state & MPTCL_STATE_LOADED) {
				fprintf(mediaPtr->mpcmd, "pause\n");
				if (mediaPtr->state & MPTCL_STATE_PAUSED) { 
					mediaPtr->state = MPTCL_STATE_LOADED ;
				} else {
					mediaPtr->state =  MPTCL_STATE_LOADED | MPTCL_STATE_PAUSED ;
				};
				/* flush */
				fflush(mediaPtr->mpcmd);
			};
			break ;
		case MPTCL_STOP : 
			if (mediaPtr->state == MPTCL_STATE_LOADED) {
				/* quit mplayer */
				fprintf(mediaPtr->mpcmd, "quit\n");
				fflush(mediaPtr->mpcmd);
				pclose(mediaPtr->mpcmd); 
				/* reinit media */
				mediaPtr->state = MPTCL_STATE_VOID;
				mediaPtr->mpwin	= 0;
				/* flush */
				fflush(mediaPtr->mpcmd);
			};
			break ;
		case MPTCL_VOLUME : 
			if (argv[2] == NULL) break ;
			if (mediaPtr->state == MPTCL_STATE_LOADED) {
				index = MPTcl_GetEqu(argv[2], mediaEqu);
				if (index == MPTCL_UP)
			 		fprintf(mediaPtr->mpcmd, "volume >\n");
				if (index == MPTCL_DOWN)
					fprintf(mediaPtr->mpcmd, "volume <\n");
			};
			break ;
		case MPTCL_MPWIN_RESIZE :
			if (mediaPtr->state == MPTCL_STATE_LOADED) {
				XGetGeometry(mediaPtr->display, mediaPtr->mediawin->window, &win, &x, &y, &w, &h, &dw, &dph);
				XResizeWindow(mediaPtr->display, mediaPtr->mpwin, w, h);
				XMoveWindow(mediaPtr->display, mediaPtr->mpwin, 0, 0);	
			}
			break;
		case MPTCL_MPWIN_SET :
			if (mediaPtr->state & MPTCL_STATE_LOADNOTIFY) {
				/* full screen asked */
				MediaRemote(mediaPtr, 0x29);MediaRemote(mediaPtr, 0x29);
				/* rename MPlayer win */
				XStoreName(mediaPtr->display, mediaPtr->mpwin, mediaPtr->mediawinName);
				/* reparent Mplayer win */
				XReparentWindow(mediaPtr->display, mediaPtr->mpwin, mediaPtr->mediawin->window, 0, 0);
				/* OK Loaded */
				mediaPtr->state = MPTCL_STATE_LOADED;
			};
			break;
		default :
			error = TCL_ERROR;
	};

	/* return error code */
	return error;
};

/*
 *--------------------------------------------------------------
 *
 * MediaWidgetConfigure --
 *
 * 	store configuration and configure Media widget
 *
 * Results:
 *	A standard Tcl result.
 *
 *--------------------------------------------------------------
 */
static int
MediaWidgetConfigure(interp,  mediaPtr, argc, argv)
    Tcl_Interp  *interp;         	/* Used for error reporting. */
    Media 	*mediaPtr;   		/* Information about widget. */
    int 	argc;                   /* Number of valid entries in argv. */
    char 	**argv;                	/* Arguments. */

{
	int  i		= 0;			/* index of args */
	int index 	= 0;			/* index of option */
	char *args;				/* string of not used argv */
	int timming	= 200;			
	char *tclStr;	
	int error	= TCL_OK;

	/* string of args */
	args 	= (char *) malloc (1024);
	args[0] = 0x00;

	/* string for tcl exec */
	tclStr	= (char *) malloc (1024);
	

	/* stock values in register */
	for  ( i = 0 ; i<argc ; i++ ) {
		index = MPTcl_GetEqu(argv[i], mediaEqu);
		i++;
		if (argv[i]==NULL) {	 
	 		sprintf(interp->result,"movie configure %s ?value?!", argv[i]);
	 		return TCL_ERROR ;
		};

		switch (index) {
			case MPTCL_BACKGROUND :
				mediaPtr->background	= atoi(argv[i]); 
				break;

			case MPTCL_DVD	:
				mediaPtr->file		= (char *) NULL;
				mediaPtr->dvd 		= atoi(argv[i]); 
				mediaPtr->state 	|= MPTCL_STATE_LOADNOTIFY;
				sprintf(tclStr, "mplayer -zoom -slave -vo x11 -dvd %d", mediaPtr->dvd);
				timming 		= 500;
				break;

			case MPTCL_FILE	:
				mediaPtr->dvd		= 0;
				mediaPtr->file		= argv[i]; 
				mediaPtr->state 	|= MPTCL_STATE_LOADNOTIFY;
				sprintf(tclStr, "mplayer -zoom -slave -vo x11 %s", mediaPtr->file);
				break;

			case MPTCL_HEIGHT : 
				mediaPtr->height 	= atoi(argv[i])	; 
				break;

			case MPTCL_SCALE : 
				mediaPtr->scale 	= atoi(argv[i])	; 
				break;

			case MPTCL_VOLUME : 
				mediaPtr->volume	= atoi(argv[i]);
				if (mediaPtr->volume<0) {mediaPtr->volume = 0 ;}; 
				break;
			case MPTCL_WIDTH : 
				mediaPtr->width 	= atoi(argv[i])	;
				break;

			default : 
				sprintf(args, "%s %s %s", args, argv[i-1], argv[i]); 
				break;
		};
	
	};

	/* Another Args Not Used */ 
	free(mediaPtr->args);
	mediaPtr->args 	=	(char *) malloc (strlen(args)+1);
	strncpy(mediaPtr->args,args, strlen(args)+1);	
	
	/* LOAD MPLAYER */
	if (mediaPtr->state & MPTCL_STATE_LOADNOTIFY && !(mediaPtr->state & MPTCL_STATE_LOADED)) {
			/* open mplayer */
			mediaPtr->mpcmd	 = popen(tclStr, "w");
			/* find mplayer win */
			for ( i = 0 ; (mediaPtr->mpwin == 0) && (i <=200) ; i++ ) {
				mediaPtr->mpwin = XFindWindowId(mediaPtr->display, mediaPtr->root, "MPlayer");
				printf("Try to find Mplayer window n%d\n", i);
			};
			if ( mediaPtr->mpwin == 0) {
				sprintf(interp->result, "Can't open MPlayer!");
				mediaPtr->state = MPTCL_STATE_VOID;
				error = TCL_ERROR;
			} else {
				/* Test (this test is very important) and wait 0.1s set mpwin */
				sprintf(tclStr,"after 500 { %s _MPWIN_SET_ ; after %d { %s _MPWIN_RESIZE_ ; bind %s  <Configure> { %s _MPWIN_RESIZE_ } } } ; winfo id %s",
					mediaPtr->mediawinName, timming, mediaPtr->mediawinName, mediaPtr->mediawinName, mediaPtr->mediawinName, mediaPtr->mediawinName);
				if ((error = Tcl_Eval(interp, tclStr)) != TCL_OK ) sprintf(interp->result, "Can't init X target windows!");
			};
	};	

	/* Geometry */
	Tk_GeometryRequest((Tk_Window) mediaPtr->mediawin, mediaPtr->width * mediaPtr->scale, mediaPtr->height * mediaPtr->scale);
	/* Default color */
	Tk_SetWindowBackground((Tk_Window) mediaPtr->mediawin, mediaPtr->background);

	/* free args */
	free (args);
	
	/* free tclStr */
	free(tclStr);

	/* return error */
	return error ;
};

/*
 *--------------------------------------------------------------
 *
 * MediaWidgetDelete --
 *
 *	called by widget command when destroy
 *
 *--------------------------------------------------------------
 */
static void
MediaWidgetDelete(clientData)
    ClientData clientData;	/* Pointer to widget record for widget. */
{
    	Media 		*mediaPtr 		= (Media *) clientData;
 	Tk_Window	 mediawin 		= (Tk_Window)  mediaPtr->mediawin;
    	Tcl_Interp 	*interp;         	/* Used for error reporting. */

	
	/* quit mplayer */
	fprintf(mediaPtr->mpcmd, "quit\n");
	fflush(mediaPtr->mpcmd);
	pclose(mediaPtr->mpcmd);

	/* destroy wtkwin */
	if (mediawin != NULL) {
		mediaPtr->mediawin = NULL;
		Tk_DestroyWindow(mediawin);
    	};

	free(mediaPtr);
};

/*
 *--------------------------------------------------------------
 *
 * XFindWindowId --
 *
 * 	return window X11 Id searched by name 
 *
 * Results:
 *	window Id
 * 
 *--------------------------------------------------------------
 */

static Window
XFindWindowId(dpy, top, name)
     Display *dpy;
     Window top;
     char *name;
{
        Window *children, dummy;
        unsigned int nchildren;
        int i;
        Window w=0;
	Window wout=0;
        char *window_name;

        if (XFetchName(dpy, top, &window_name) && !strcmp(window_name, name))
          return(top);

        if (!XQueryTree(dpy, top, &dummy, &dummy, &children, &nchildren))
          return(0);

        for (i=0; i<nchildren; i++) {
                w = XFindWindowId(dpy, children[i], name);
                if (w) wout=w;
        }
        if (children) XFree ((char *)children);
        return(wout);
}


/*
 *--------------------------------------------------------------
 *
 * MediaRemote --
 *
 * 	called by widget cmd for control the current media
 *
 * Results:
 *	A standard Tcl result.
 *
 *--------------------------------------------------------------
 */
static int
MediaRemote(mediaPtr, key)
    	Media	*mediaPtr;	/* Media pointer. */
    	unsigned int	key;			/* MPlayer key code */

{
   	XEvent			event;			/* new event */
	XKeyEvent		keyevent;		/* new key event */
	XAnyEvent		anyevent;	

	/* init event and keyevent */
	anyevent.display 	= keyevent.display 	= mediaPtr->display;
	anyevent.window		= keyevent.window	= mediaPtr->mpwin;
	anyevent.send_event	= keyevent.send_event 	= True;
	keyevent.root		= RootWindow(mediaPtr->display, DefaultScreen(mediaPtr->display));
	keyevent.keycode 	= key;
	keyevent.state 		= ( keyevent.keycode > 0xFF ) ?  0xFF : 0x00;
	event.type 		= anyevent.type		=	keyevent.type	 = KeyPress;
	event.xkey		= keyevent;
	event.xany		= anyevent;

	/* send key event */
	XSendEvent(mediaPtr->display, mediaPtr->mpwin, True, KeyPressMask, &event);
	
	/* OK */
	return TCL_OK;
};

















